/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "cmdimageplane.h"
#include "graleshellinstance.h"
#include "graleshellimageplane.h"
#include <shellp/iosystem.h>
#include <grale/imageplanegnuplotwriter.h>
#include <grale/imageplanepngwriter.h>
#include <grale/imageplaneasciiwriter.h>
#include <grale/imageplanefitswriter.h>
#include <grale/imagesdata.h>
#include <grale/gravitationallens.h>
#include <grale/triangle2d.h>
#include <grale/imageplane.h>
#include <grale/sourceplane.h>
#include <enut/ipv4address.h>
#include <stdio.h>

#include "debugnew.h"

using namespace shellp;

CmdImagePlane::CmdImagePlane(const std::string &cmdname) : ShellCommand(cmdname)
{
	x1arg = new ShellCmdRealArg("leftX");
	x1arg->setUnitType(UNIT_ANGULAR);
	x1arg->setDescription("X coordinate of the left side of the image plane.");
	addArgument(x1arg);
	y1arg = new ShellCmdRealArg("bottomY");
	y1arg->setUnitType(UNIT_ANGULAR);
	y1arg->setDescription("Y coordinate of the bottom of the image plane.");
	addArgument(y1arg);
	x2arg = new ShellCmdRealArg("rightX");
	x2arg->setUnitType(UNIT_ANGULAR);
	x2arg->setDescription("X coordinate of the right side of the image plane.");
	addArgument(x2arg);
	y2arg = new ShellCmdRealArg("topY");
	y2arg->setUnitType(UNIT_ANGULAR);
	y2arg->setDescription("Y coordinate of the top of the image plane.");
	addArgument(y2arg);
	numxarg = new ShellCmdIntArg("numX","512");
	numxarg->setMin(16);
	numxarg->setDescription("Number of points to sample the image plane in the X direction.");
	addArgument(numxarg);
	numyarg = new ShellCmdIntArg("numY","512");
	numyarg->setMin(16);
	numyarg->setDescription("Number of points to sample the image plane in the Y direction.");
	addArgument(numyarg);
}

CmdImagePlane::~CmdImagePlane()
{
	delete x1arg;
	delete x2arg;
	delete y1arg;
	delete y2arg;
	delete numxarg;
	delete numyarg;
}

CmdImagePlaneLocal::CmdImagePlaneLocal(const std::string &cmdname) : CmdImagePlane(cmdname)
{
	setDescription("Create image plane based upon the current lens and source plane. Calculations are done locally.");
}

CmdImagePlaneLocal::~CmdImagePlaneLocal()
{
}

bool CmdImagePlaneLocal::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens;
	grale::SourcePlane *srcplane;
	
	if ((lens = gsinst->GetLens()) == 0)
	{
		setErrorString("A lens must have been created first");
		return false;
	}
	if ((srcplane = gsinst->GetSourcePlane()) == 0)
	{
		setErrorString("A source plane must have been created first");
		return false;
	}

	grale::Vector2D<double> bottomleft(GetX1(),GetY1());
	grale::Vector2D<double> topright(GetX2(),GetY2());
	
	GraleShellImagePlane *imgplane = new GraleShellImagePlane(lens,srcplane);

	if (!imgplane->init(bottomleft,topright,GetNumX(),GetNumY(), gsinst->GetOpenCLKernel()))
	{
		setErrorString(std::string("Couldn't initialize the image plane: ") + imgplane->getErrorString());
		delete imgplane;
		return false;
	}
	
	gsinst->SetImagePlane(imgplane);
	
	return true;
}

CmdImagePlaneRemote::CmdImagePlaneRemote(const std::string &cmdname) : CmdImagePlane(cmdname)
{
	ipstrarg = new ShellCmdStringArg("IPaddress");
	ipstrarg->setDescription("IP address of the rendering server.");
	addArgument(ipstrarg);
	portarg = new ShellCmdIntArg("port");
	portarg->setDescription("Port number of the rendering server.");
	portarg->setMin(1);
	portarg->setMax(65535);
	addArgument(portarg);
	setDescription("Create image plane based upon the current lens and source plane. Calculations are done using a remote rendering server.");
}

CmdImagePlaneRemote::~CmdImagePlaneRemote()
{
	delete ipstrarg;
	delete portarg;
}

bool CmdImagePlaneRemote::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens;
	grale::SourcePlane *srcplane;
	
	if ((lens = gsinst->GetLens()) == 0)
	{
		setErrorString("A lens must have been created first");
		return false;
	}
	if ((srcplane = gsinst->GetSourcePlane()) == 0)
	{
		setErrorString("A source plane must have been created first");
		return false;
	}

	grale::Vector2D<double> bottomleft(GetX1(),GetY1());
	grale::Vector2D<double> topright(GetX2(),GetY2());
	
	nut::IPv4Address address;

	if (!address.setAddress(ipstrarg->getValue()))
	{
		setErrorString("An invalid IP address was entered");
		return false;
	}
	
	uint16_t port = (uint16_t)(portarg->getValue());
	
	GraleShellImagePlane *imgplane = new GraleShellImagePlane(lens,srcplane);

	if (!imgplane->init(bottomleft, topright, GetNumX(), GetNumY(), address, port))
	{
		setErrorString(std::string("Couldn't initialize the image plane: ") + imgplane->getErrorString());
		delete imgplane;
		return false;
	}
	
	gsinst->SetImagePlane(imgplane);
	
	return true;
}

CmdImagePlanePlotGnuplot::CmdImagePlanePlotGnuplot(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Output file name.");
	addArgument(fnamearg);
	imgarg = new ShellCmdBoolArg("images","yes");
	imgarg->setDescription("Flag indicating if images should be drawn.");
	addArgument(imgarg);
	srcarg = new ShellCmdBoolArg("sources","yes");
	srcarg->setDescription("Flag indicating if sources should be drawn.");
	addArgument(srcarg);
	critarg = new ShellCmdBoolArg("crit","yes");
	critarg->setDescription("Flag indicating if critical lines should be drawn.");
	addArgument(critarg);
	caustarg = new ShellCmdBoolArg("caust","yes");
	caustarg->setDescription("Flag indicating if caustics should be drawn.");
	addArgument(caustarg);
	xlabelarg = new ShellCmdStringArg("xlabel","X");
	xlabelarg->setDescription("Label for the x-axis.");
	addArgument(xlabelarg);
	ylabelarg = new ShellCmdStringArg("ylabel","Y");
	ylabelarg->setDescription("Label for the y-axis.");
	addArgument(ylabelarg);
	colorarg = new ShellCmdBoolArg("color","no");
	colorarg->setDescription("Flag indicating if color output should be used.");
	addArgument(colorarg);
	fntsizearg = new ShellCmdIntArg("fontsize","18");
	fntsizearg->setMin(1);
	fntsizearg->setDescription("Font point size.");
	addArgument(fntsizearg);
	gridarg = new ShellCmdBoolArg("grid","no");
	gridarg->setDescription("Flag indicating if a grid should be drawn.");
	addArgument(gridarg);
	m_pFlipXArg = new ShellCmdBoolArg("flipx");
	m_pFlipXArg->setDefault("no");
	m_pFlipXArg->setDescription("If set, the x-axis on the plot is reversed.");
	addArgument(m_pFlipXArg);
	m_pSubSample = new ShellCmdIntArg("subsample");
	m_pSubSample->setMin(1);
	m_pSubSample->setDescription("Number of points which each pixel should be subsampled.");
	m_pSubSample->setDefault("10");
	addArgument(m_pSubSample);
	setDescription("Create a gnuplot file of the image plane.");
}
	
CmdImagePlanePlotGnuplot::~CmdImagePlanePlotGnuplot()
{
	delete fnamearg;
	delete imgarg;
	delete srcarg;
	delete critarg;
	delete caustarg;
	delete xlabelarg;
	delete ylabelarg;
	delete colorarg;
	delete fntsizearg;
	delete gridarg;
	delete m_pSubSample;
	delete m_pFlipXArg;
}

bool CmdImagePlanePlotGnuplot::execute()
{	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *imgplane = gsinst->GetImagePlane();

	if (imgplane == 0)
	{
		setErrorString("No image plane has been created yet.");
		return false;
	}
	
	grale::ImagePlaneGnuPlotWriter w(fnamearg->getValue(),imgarg->getValue(),srcarg->getValue(),critarg->getValue(),caustarg->getValue(),gsinst->GetAngularUnit(),m_pFlipXArg->getValue());
	
	w.setXLabel(xlabelarg->getValue());
	w.setYLabel(ylabelarg->getValue());
	w.setGrid(gridarg->getValue());
	w.setFontSize(fntsizearg->getValue());
	w.setColor(colorarg->getValue());
	
	if (!w.write(imgplane, m_pSubSample->getValue()))
	{
		setErrorString(std::string("Couldn't plot image plane: ") + w.getErrorString());
		return false;
	}
		
	return true;
}

CmdImagePlanePlotPNG::CmdImagePlanePlotPNG(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Output file name.");
	addArgument(fnamearg);
	imgarg = new ShellCmdBoolArg("images","yes");
	imgarg->setDescription("Flag indicating if images should be drawn.");
	addArgument(imgarg);
	srcarg = new ShellCmdBoolArg("sources","yes");
	srcarg->setDescription("Flag indicating if sources should be drawn.");
	addArgument(srcarg);
	critarg = new ShellCmdBoolArg("crit","yes");
	critarg->setDescription("Flag indicating if critical lines should be drawn.");
	addArgument(critarg);
	caustarg = new ShellCmdBoolArg("caust","yes");
	caustarg->setDescription("Flag indicating if caustics should be drawn.");
	addArgument(caustarg);
	m_pSubSample = new ShellCmdIntArg("subsample");
	m_pSubSample->setMin(1);
	m_pSubSample->setDescription("Number of points which each pixel should be subsampled.");
	m_pSubSample->setDefault("10");
	addArgument(m_pSubSample);

	setDescription("Create a PNG file of the image plane.");
}
	
CmdImagePlanePlotPNG::~CmdImagePlanePlotPNG()
{
	delete fnamearg;
	delete imgarg;
	delete srcarg;
	delete critarg;
	delete caustarg;
	delete m_pSubSample;
}

bool CmdImagePlanePlotPNG::execute()
{	
#ifdef GRALECONFIG_SUPPORT_QT4
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *imgplane = gsinst->GetImagePlane();

	if (imgplane == 0)
	{
		setErrorString("No image plane has been created yet.");
		return false;
	}
	
	grale::ImagePlanePNGWriter w(fnamearg->getValue(),imgarg->getValue(),srcarg->getValue(),critarg->getValue(),caustarg->getValue());
	
	if (!w.write(imgplane, m_pSubSample->getValue()))
	{
		setErrorString(std::string("Couldn't plot image plane: ") + w.getErrorString());
		return false;
	}
	
	return true;
#else
	setErrorString("No PNG support was compiled in.");
	return false;
#endif // GRALECONFIG_SUPPORT_QT4
}

CmdImagePlanePlotAscii::CmdImagePlanePlotAscii(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Output file name.");
	addArgument(fnamearg);
	imgarg = new ShellCmdBoolArg("images","yes");
	imgarg->setDescription("Flag indicating if images should be drawn.");
	addArgument(imgarg);
	srcarg = new ShellCmdBoolArg("sources","no");
	srcarg->setDescription("Flag indicating if sources should be drawn.");
	addArgument(srcarg);
	critarg = new ShellCmdBoolArg("crit","no");
	critarg->setDescription("Flag indicating if critical lines should be drawn.");
	addArgument(critarg);
	caustarg = new ShellCmdBoolArg("caust","no");
	caustarg->setDescription("Flag indicating if caustics should be drawn.");
	addArgument(caustarg);
	m_pSubSample = new ShellCmdIntArg("subsample");
	m_pSubSample->setMin(1);
	m_pSubSample->setDescription("Number of points with which each pixel should be subsampled.");
	m_pSubSample->setDefault("10");
	addArgument(m_pSubSample);

	setDescription("Create a raw ASCII file with data of the image plane.");
}
	
CmdImagePlanePlotAscii::~CmdImagePlanePlotAscii()
{
	delete fnamearg;
	delete imgarg;
	delete srcarg;
	delete critarg;
	delete caustarg;
	delete m_pSubSample;
}

bool CmdImagePlanePlotAscii::execute()
{	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *imgplane = gsinst->GetImagePlane();

	if (imgplane == 0)
	{
		setErrorString("No image plane has been created yet.");
		return false;
	}

	FILE *f;	
	
	f = fopen(fnamearg->getValue().c_str(),"wt");
	if (f == 0)
	{
		setErrorString("Can't open file for writing");
		return false;
	}
	
	grale::ImagePlaneASCIIWriter w(f,imgarg->getValue(),srcarg->getValue(),critarg->getValue(),caustarg->getValue(),gsinst->GetAngularUnit());
	
	if (!w.write(imgplane, m_pSubSample->getValue()))
	{
		setErrorString(std::string("Couldn't plot image plane: ") + w.getErrorString());
		return false;
	}

	fclose(f);
	
	return true;
}

CmdImagePlanePlotFITS::CmdImagePlanePlotFITS(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Output file name.");
	addArgument(fnamearg);
	m_pSubSample = new ShellCmdIntArg("subsample");
	m_pSubSample->setMin(1);
	m_pSubSample->setDescription("Number of points with which each pixel should be subsampled.");
	m_pSubSample->setDefault("10");
	addArgument(m_pSubSample);
	m_pPlotImages = new ShellCmdBoolArg("plotimages");
	m_pPlotImages->setDefault("yes");
	m_pPlotImages->setDescription("Flag indicating if the images or the sources should be plotted.");
	addArgument(m_pPlotImages);

	setDescription("Create a FITS file of the images in the image plane.");
}
	
CmdImagePlanePlotFITS::~CmdImagePlanePlotFITS()
{
	delete fnamearg;
	delete m_pSubSample;
	delete m_pPlotImages;
}

bool CmdImagePlanePlotFITS::execute()
{	
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *imgplane = gsinst->GetImagePlane();

	if (imgplane == 0)
	{
		setErrorString("No image plane has been created yet.");
		return false;
	}

	grale::ImagePlaneFITSWriter w(fnamearg->getValue(), gsinst->GetAngularUnit(), m_pPlotImages->getValue());
	
	if (!w.write(imgplane, m_pSubSample->getValue()))
	{
		setErrorString(std::string("Couldn't plot image plane: ") + w.getErrorString());
		return false;
	}

	return true;
#else 
	setErrorString("No cfitsio support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}

CmdImagePlanePlotMagnification::CmdImagePlanePlotMagnification(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Output file name.");
	addArgument(fnamearg);
	setDescription("Write magnification info to a file.");
}

CmdImagePlanePlotMagnification::~CmdImagePlanePlotMagnification()
{
	delete fnamearg;
}

bool CmdImagePlanePlotMagnification::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *imgplane = gsinst->GetImagePlane();

	if (imgplane == 0)
	{
		setErrorString("No image plane has been created yet.");
		return false;
	}

	FILE *f = fopen(fnamearg->getValue().c_str(),"wt");
	if (f == 0)
	{
		setErrorString("Couldn't open speficied file");
		return false;
	}

	int numx = imgplane->getNumXPoints();
	int numy = imgplane->getNumYPoints();

	for (int y = 0 ; y < numy ; y++)
	{
		for (int x = 0 ; x < numx ; x++)
		{
			double val = 1.0/ABS(imgplane->getInverseMagnification(x,y));
			grale::Vector2D<double> pos = imgplane->getIndexCoordinate(x,y)/gsinst->GetAngularUnit();
			fprintf(f,"%30.30g %30.30g %30.30g\n",(double)pos.getX(),(double)pos.getY(),(double)val);
		}
		fprintf(f,"\n");
	}

	fclose(f);

	return true;
}

CmdImagePlaneLensPlane::CmdImagePlaneLensPlane(const std::string &cmdname) : ShellCommand(cmdname)
{
	setDescription("Build an image plane from an existing lens plane");
}

CmdImagePlaneLensPlane::~CmdImagePlaneLensPlane()
{
}

bool CmdImagePlaneLensPlane::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::LensPlane *lensplane;
	grale::SourcePlane *srcplane;
	
	if ((lensplane = gsinst->GetLensPlane()) == 0)
	{
		setErrorString("A lens plane must have been created first");
		return false;
	}
	if ((srcplane = gsinst->GetSourcePlane()) == 0)
	{
		setErrorString("A source plane must have been created first");
		return false;
	}

	GraleShellImagePlane *imgplane = new GraleShellImagePlane(0,srcplane);

	if (!imgplane->init(lensplane))
	{
		setErrorString(std::string("Couldn't initialize the image plane: ") + imgplane->getErrorString());
		delete imgplane;
		return false;
	}
	
	gsinst->SetImagePlane(imgplane);
	
	return true;
}

CmdImagePlaneTraceBase::CmdImagePlaneTraceBase(const std::string &cmdName) : ShellCommand(cmdName)
{
}

CmdImagePlaneTraceBase::~CmdImagePlaneTraceBase()
{
}

void CmdImagePlaneTraceBase::ListPoint(IOSystem *pIOSys, grale::Vector2D<double> point)
{
	char str[1024];

//	snprintf(str, 1023, "  %3.10g %3.10g", point.getX(), point.getY());
	sprintf(str, "  %3.10g %3.10g", point.getX(), point.getY());
	str[1023] = 0;

	pIOSys->writeOutputLine(str);
}

void CmdImagePlaneTraceBase::ListPoint(IOSystem *pIOSys, grale::Vector2D<double> point, double magnification)
{
	char str[1024];

	//snprintf(str, 1023, "  %3.10g %3.10g %3.10g", point.getX(), point.getY(), magnification);
	sprintf(str, "  %3.10g %3.10g %3.10g", point.getX(), point.getY(), magnification);
	str[1023] = 0;

	pIOSys->writeOutputLine(str);
}

CmdImagePlaneTraceBeta::CmdImagePlaneTraceBeta(const std::string &cmdName) : CmdImagePlaneTraceBase(cmdName)
{
	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pXArg->setDescription("X component of the source plane point.");
	addArgument(m_pXArg);

	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_pYArg->setDescription("Y component of the source plane point.");
	addArgument(m_pYArg);

	m_pNumIterationsArg = new ShellCmdIntArg("iterations");
	m_pNumIterationsArg->setMin(1);
	m_pNumIterationsArg->setDefault("4");
	m_pNumIterationsArg->setDescription("The number of iterations when looking for a refinement in the image plane position.");
	addArgument(m_pNumIterationsArg);

	m_pMagnificationArg = new ShellCmdBoolArg("magnification");
	m_pMagnificationArg->setDefault("true");
	m_pMagnificationArg->setDescription("If set, magnification values will be listed as well.");
	addArgument(m_pMagnificationArg);
	
	setDescription("Finds the image plane points corresponding to the specified source plane point.");
}

CmdImagePlaneTraceBeta::~CmdImagePlaneTraceBeta()
{
	delete m_pXArg;
	delete m_pYArg;
	delete m_pNumIterationsArg;
	delete m_pMagnificationArg;
}

bool CmdImagePlaneTraceBeta::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInstance->getIOSystem();
	grale::ImagePlane *pImagePlane = pInstance->GetImagePlane();
	
	if (pImagePlane == 0)
	{
		setErrorString("No image plane has been created yet");
		return false;
	}

	int width = pImagePlane->getNumXPoints();
	int height = pImagePlane->getNumYPoints();
	grale::Vector2D<double> beta(m_pXArg->getValue(), m_pYArg->getValue());
	std::list<grale::Vector2D<double> > thetaPoints;
	std::list<grale::Vector2D<double> >::const_iterator it;

	if (!pImagePlane->traceBeta(beta, thetaPoints, m_pNumIterationsArg->getValue()))
	{
		setErrorString(std::string("Unable to trace specified position: ") + pImagePlane->getErrorString());
		return false;
	}

	for (it = thetaPoints.begin() ; it != thetaPoints.end() ; it++)
	{
		grale::Vector2D<double> theta = (*it);

		if (m_pMagnificationArg)
			ListPoint(pIOSys, theta/pInstance->GetAngularUnit(), 1.0/pImagePlane->getInverseMagnification(theta));
		else
			ListPoint(pIOSys, theta/pInstance->GetAngularUnit());
	}
	
	return true;
}

CmdImagePlaneTraceTheta::CmdImagePlaneTraceTheta(const std::string &cmdName) : CmdImagePlaneTraceBase(cmdName)
{
	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pXArg->setDescription("X component of the image plane point.");
	addArgument(m_pXArg);

	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_pYArg->setDescription("Y component of the image plane point.");
	addArgument(m_pYArg);

	setDescription("Finds the source plane points corresponding to the specified image plane point.");
}

CmdImagePlaneTraceTheta::~CmdImagePlaneTraceTheta()
{
	delete m_pXArg;
	delete m_pYArg;
}

bool CmdImagePlaneTraceTheta::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInstance->getIOSystem();
	grale::ImagePlane *pImagePlane = pInstance->GetImagePlane();
	
	if (pImagePlane == 0)
	{
		setErrorString("No image plane has been created yet");
		return false;
	}

	grale::Vector2D<double> theta(m_pXArg->getValue(), m_pYArg->getValue());
	grale::Vector2D<double> beta;
	
	if (!pImagePlane->traceTheta(theta, &beta))
	{
		setErrorString(pImagePlane->getErrorString());
		return false;
	}
	
	ListPoint(pIOSys, beta/pInstance->GetAngularUnit());
	
	return true;
}

CmdImagePlaneTimeDelay::CmdImagePlaneTimeDelay(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pLensRedshiftArg = new ShellCmdRealArg("z_lens");
	m_pLensRedshiftArg->setMin(0);
	m_pLensRedshiftArg->setDescription("Redshift of the lens");
	addArgument(m_pLensRedshiftArg);

	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pXArg->setDescription("X component of the source plane point.");
	addArgument(m_pXArg);

	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_pYArg->setDescription("Y component of the source plane point.");
	addArgument(m_pYArg);

	m_pNumIterationsArg = new ShellCmdIntArg("iterations");
	m_pNumIterationsArg->setMin(1);
	m_pNumIterationsArg->setDefault("4");
	m_pNumIterationsArg->setDescription("The number of iterations when looking for a refinement in the image plane position.");
	addArgument(m_pNumIterationsArg);
	
	setDescription("Finds the image plane points corresponding to the specified source plane point.");
}

CmdImagePlaneTimeDelay::~CmdImagePlaneTimeDelay()
{
	delete m_pLensRedshiftArg;
	delete m_pXArg;
	delete m_pYArg;
	delete m_pNumIterationsArg;
}

bool CmdImagePlaneTimeDelay::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInstance->getIOSystem();
	grale::ImagePlane *pImagePlane = pInstance->GetImagePlane();
	
	if (pImagePlane == 0)
	{
		setErrorString("No image plane has been created yet");
		return false;
	}

	int width = pImagePlane->getNumXPoints();
	int height = pImagePlane->getNumYPoints();
	grale::Vector2D<double> beta(m_pXArg->getValue(), m_pYArg->getValue());
	std::list<grale::Vector2D<double> > thetaPoints;
	std::list<grale::Vector2D<double> >::const_iterator it;
	std::list<double> timeDelays;
	std::list<double>::const_iterator it2;

	if (!pImagePlane->traceBeta(beta, thetaPoints, m_pNumIterationsArg->getValue()))
	{
		setErrorString(std::string("Unable to trace specified position: ") + pImagePlane->getErrorString());
		return false;
	}

	const grale::GravitationalLens *pLens = pImagePlane->getLens();
	double D_s = pImagePlane->getSourcePlane()->getD_s();
	double D_ds = pImagePlane->getSourcePlane()->getD_ds();
	double minDelay = 1e30;

	for (it = thetaPoints.begin() ; it != thetaPoints.end() ; it++)
	{
		double timeDelay = 0;

		if (!pLens->getTimeDelay(m_pLensRedshiftArg->getValue(), D_s, D_ds, *it, beta, &timeDelay))
		{
			setErrorString(std::string("Unable to get time delay info: ") + pLens->getErrorString());
			return false;
		}

		timeDelays.push_back(timeDelay);
		if (minDelay > timeDelay)
			minDelay = timeDelay;
	}

	for (it = thetaPoints.begin(), it2 = timeDelays.begin() ; it != thetaPoints.end() ; it++, it2++)
	{
		grale::Vector2D<double> theta = *it;
		double timeDelay = *it2;

		ListPoint(pIOSys, theta/pInstance->GetAngularUnit(), timeDelay-minDelay);
	}

	return true;
}

void CmdImagePlaneTimeDelay::ListPoint(IOSystem *pIOSys, grale::Vector2D<double> point, double timeDelay)
{
	char str[1024];

	//snprintf(str, 1023, "  %3.10g %3.10g: %g days", point.getX(), point.getY(), timeDelay/(60*60*24));
	sprintf(str, "  %3.10g %3.10g: %g days", point.getX(), point.getY(), timeDelay/(60*60*24));
	str[1023] = 0;

	pIOSys->writeOutputLine(str);
}

CmdImagePlaneTraceImgData::CmdImagePlaneTraceImgData(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pFileNameArg = new ShellCmdStringArg("filename");
	m_pFileNameArg->setDescription("Name of the file to which the back-projected images data should be written.");
	addArgument(m_pFileNameArg);

	setDescription("Project the current images data back onto the source plane associated with the "
	               "current image plane. Because the lens equation will be approximated using the "
		       "data stored in the image plane instance, the results will be approximations "
		       "as well.");
}

CmdImagePlaneTraceImgData::~CmdImagePlaneTraceImgData()
{
	delete m_pFileNameArg;
}

bool CmdImagePlaneTraceImgData::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *pImgPlane = pGSInst->GetImagePlane();
	grale::ImagesDataExtended *pImgData = pGSInst->GetImagesData();

	if (pImgPlane == 0)
	{
		setErrorString("No image plane has been created yet");
		return false;
	}

	if (pImgData == 0)
	{
		setErrorString("No images data set has been created");
		return false;
	}

	FILE *pFile = fopen(m_pFileNameArg->getValue().c_str(), "wt");
	if (pFile == 0)
	{
		setErrorString("Can't open specified file for writing");
		return false;
	}

	int numImages = pImgData->getNumberOfImages();

	for (int i = 0 ; i < numImages ; i++)
	{
		int numImagePoints = pImgData->getNumberOfImagePoints(i);

		for (int j = 0 ; j < numImagePoints ; j++)
		{
			grale::Vector2D<double> theta = pImgData->getImagePointPosition(i, j);
			grale::Vector2D<double> beta;
			double intens = 0;

			if (pImgData->hasIntensities())
				intens = pImgData->getImagePointIntensity(i, j);

			if (!pImgPlane->traceThetaApprox(theta, &beta))
			{
				fclose(pFile);
				setErrorString(std::string("Error projecting images data onto source plane using calculated image plane data: ") + pImgPlane->getErrorString());
				return false;
			}

			double invMag = pImgPlane->getInverseMagnificationApprox(theta);

			fprintf(pFile, "%g %g %g\n", (double)(beta.getX()/pGSInst->GetAngularUnit()), 
					(double)(beta.getY()/pGSInst->GetAngularUnit()),
					intens/invMag);
		}
		fprintf(pFile, "\n");
	}

	fclose(pFile);
	return true;
}

