/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleshellinstance.h"
#include "graleshellgaparams.h"
#include <shellp/iosystem.h>
#include <grale/gravitationallens.h>
#include <grale/imageplane.h>
#include <grale/lensplane.h>
#include <grale/sourceplane.h>
#include <grale/sourceimage.h>
#include <grale/constants.h>
#include <grale/imagesdata.h>
#include <grale/grid.h>
#include <grale/cosmology.h>
#include <grale/openclkernel.h>
#include <grale/randomnumbergenerator.h>
#include <grale/uniformdistribution.h>
#include <grale/gaussiandistribution.h>
#include <iostream>
#include <stdlib.h>

#include "debugnew.h"

double CalculateAngDiamDist1(double z)
{
	GraleShellInstance *pInst = (GraleShellInstance *)shellp::ShellInstance::getInstance();
	grale::Cosmology *pCosmology = pInst->GetCosmology();
	double unit = pInst->GetDistanceUnit();
	double dist = pCosmology->getAngularDiameterDistance(0, z);

	return dist/unit;
}

double CalculateAngDiamDist2(double z1, double z2)
{
	GraleShellInstance *pInst = (GraleShellInstance *)shellp::ShellInstance::getInstance();
	grale::Cosmology *pCosmology = pInst->GetCosmology();
	double unit = pInst->GetDistanceUnit();

	if (z2 < z1)
	{
		double tmp = z1;

		z1 = z2;
		z2 = tmp;
	}

	double dist = pCosmology->getAngularDiameterDistance(z1, z2);

	return dist/unit;
}

double GetRandomNumberInInterval(double x0, double x1)
{
	GraleShellInstance *pInst = (GraleShellInstance *)shellp::ShellInstance::getInstance();
	const grale::RandomNumberGenerator *pRnd = pInst->GetRandomNumberGenerator();
	double min = MIN(x0, x1);
	double max = MAX(x0, x1);
	grale::UniformDistribution uniDist(min, max, pRnd);

	return uniDist.pickNumber();
}

double GetGaussianRandomNumber(double mu, double sigma)
{
	GraleShellInstance *pInst = (GraleShellInstance *)shellp::ShellInstance::getInstance();
	const grale::RandomNumberGenerator *pRnd = pInst->GetRandomNumberGenerator();
	grale::GaussianDistribution gaussDist(mu, sigma, pRnd);

	return gaussDist.pickNumber();
}

GraleShellInstance::GraleShellInstance()
{
	registerUnitName(UNIT_ANGULAR, "Angular");
	registerUnitName(UNIT_MASS, "Mass");
	registerUnitName(UNIT_DISTANCE, "Distance");

	SetAngularUnit(ANGLE_ARCSEC);
	SetMassUnit(MASS_SOLAR);
	SetDistanceUnit(DIST_MPC);
	
	defineFunction("z", CalculateAngDiamDist1);
	defineFunction("zz", CalculateAngDiamDist2);
	defineFunction("rnduni", GetRandomNumberInInterval);
	defineFunction("rndgauss", GetGaussianRandomNumber);

	lens = 0;
	imgplane = 0;
	srcplane = 0;
	imagesdata = 0;
	grid = 0;
	lensplane = 0;
	
	gaparams = new GraleShellGAParams(); // set default parameters
	m_pCosmology = new grale::Cosmology();
	m_pOpenCLKernel = 0;

	m_pRndGen = new grale::RandomNumberGenerator();
}

GraleShellInstance::~GraleShellInstance()
{
	SetLens(0);
	SetSourcePlane(0);
	SetImagePlane(0);
	SetLensPlane(0);
	SetImagesData(0);
	ClearImagesDataList();
	SetGrid(0);
	SetGAParams(0);
	SetCosmology(0);
	SetOpenCLKernel(0);
}

void GraleShellInstance::SetLens(grale::GravitationalLens *l)
{
	if (lens)
		delete lens;
	lens = l;
}

void GraleShellInstance::SetSourcePlane(grale::SourcePlane *s, bool deletesources)
{
	bool deletesrcplane = true;
	
	if (srcplane == 0)
		deletesrcplane = false;
	else
	{
		if (imgplane != 0)
		{
			grale::SourcePlane *p = imgplane->getSourcePlane();

			if (p == srcplane)
				deletesrcplane = false;
		}
	}

	if (deletesrcplane)
	{
		// ok, delete the source plane
		delete srcplane; // this also unregisters the sources;
		
		if (deletesources)
			ClearSources();
	}
	else
	{
		if (srcplane == 0)
		{
			if (deletesources)
				ClearSources();
		}
		else // srcplane is referenced in image plane
		{
			// we should create copies of all the sources so that modifying them won't
			// affect the instance in the image plane

			std::map<uint16_t, grale::SourceImage *>::const_iterator it;
			std::map<uint16_t, grale::SourceImage *> newsources;
			
			for (it = sources.begin() ; it != sources.end() ; it++)
			{
				grale::SourceImage *src = (*it).second;

				if (!deletesources)
				{
					grale::SourceImage *newsrc = src->createCopy();
					newsources[(*it).first] = newsrc; // store it
				}
			}
			sources = newsources; // if deletesources is true, newsources is empty
		}
	}

	srcplane = s;
		
	if (srcplane)
	{
		// Currently, it is possible that the srcplane itself and our local sources map
		// contain sources. We'll have to check if we can merge the two maps.

		bool mergeable = true;
		std::map<uint16_t, grale::SourceImage *> newsources = srcplane->getSources(); // NOTE: do NOT use a reference here since we'll be needing the exact state of the map at this moment
		std::map<uint16_t, grale::SourceImage *>::const_iterator it;

		for (it = newsources.begin() ; mergeable && it != newsources.end() ; it++)
		{
			if (sources.find((*it).first) != sources.end())
				mergeable = false;
		}

		if (mergeable)
		{
			// first, attach the sources in the 'sources' map
			for (it = sources.begin() ; it != sources.end() ; it++)
			{
				if (!(*it).second->attach(srcplane,(*it).first))
					unrecoverableError(std::string("ERROR: SetSourcePlane: can't attach source during merge: ") + (*it).second->getErrorString());
			}
			// add the sources in 'newsources' to the 'sources' map
			// these sources are already attached to the source plane
			for (it = newsources.begin() ; it != newsources.end() ; it++)
				sources[(*it).first] = (*it).second;
		}
		else
		{
			getIOSystem()->writeOutputLine("WARNING: deleting sources stored in the specified source plane");
			
			// The maps cannot be merged. Since we specified explicitly that the
			// sources in 'sources' should be kept, we'll clear the sources stored
			// in 'srcplane'
			for (it = newsources.begin() ; it != newsources.end() ; it++)
				delete (*it).second;
			
			// then, attach the sources in the 'sources' map
			for (it = sources.begin() ; it != sources.end() ; it++)
			{
				if (!(*it).second->attach(srcplane,(*it).first))
					unrecoverableError(std::string("ERROR: SetSourcePlane: can't attach source: ") + (*it).second->getErrorString());
			}
		}
	}

	// check parent
	
	std::map<uint16_t, grale::SourceImage *>::const_iterator it,it2;
	
	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		if ((*it).second->getParentPlane() != srcplane)
			unrecoverableError("ERROR: incorrent parent source plane");
	}

	// check for double entries
	
	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		for (it2 = sources.begin() ; it2 != sources.end() ; it2++)
		{
			if ((*it).first != (*it2).first && (*it).second == (*it2).second)
				unrecoverableError("ERROR: double entry in the sources array");
		}
	}
}

void GraleShellInstance::SetImagePlane(grale::ImagePlane *i)
{
	if (imgplane)
	{
		grale::SourcePlane *p = imgplane->getSourcePlane();

		// the sourceplane in the image plane can only be deleted if it is not the active source plane
		if (srcplane != p)
		{
			// delete the associated sources
			
			std::map<uint16_t, grale::SourceImage *> newsources = p->getSources();
			std::map<uint16_t, grale::SourceImage *>::iterator it;

			for (it = newsources.begin() ; it != newsources.end() ; it++)
				delete (*it).second;
			
			// delete the sourceplane itself
			
			delete p;
		}
		delete imgplane;
	}

	imgplane = i;
}

void GraleShellInstance::SetLensPlane(grale::LensPlane *lp)
{
	if (lensplane)
		delete lensplane;

	lensplane = lp;
}

void GraleShellInstance::SetSource(uint16_t i, grale::SourceImage *s)
{
	std::map<uint16_t, grale::SourceImage *>::iterator it;
	
	if ((it = sources.find(i)) != sources.end())
	{
		if (FindSource(s))
		{
			unrecoverableError("ERROR: source already stored in the shell instance somewhere");
		}
		delete (*it).second;
		sources.erase(it);
	}
	if (s)
		sources[i] = s;

	if (srcplane && s) // attach the source
	{
		if (!s->attach(srcplane,i))
		{
			unrecoverableError(std::string("ERROR: SetSource: can't attach source: ") + s->getErrorString());
		}
	}
}

void GraleShellInstance::ClearSources()
{
	std::map<uint16_t, grale::SourceImage *>::iterator it;

	for (it = sources.begin() ; it != sources.end() ; it++)
		delete (*it).second;
	sources.clear();
}

void GraleShellInstance::AddToImagesDataList(grale::ImagesDataExtended *imgsdata)
{
	if (imgsdata != 0)
		imagesdatalist.push_back(imgsdata);
}

void GraleShellInstance::ClearImagesDataList()
{
	std::list<grale::ImagesDataExtended *>::iterator it;

	it = imagesdatalist.begin();
	while (it != imagesdatalist.end())
	{
		grale::ImagesDataExtended *data = *it;
		
		imagesdatalist.pop_front();
		it = imagesdatalist.begin();
		
		if (!FindImagesData(data))
			delete data;
	}
}

void GraleShellInstance::SetImagesData(grale::ImagesDataExtended *imgsdata)
{
	if (imgsdata == imagesdata)
		return;
	
	if (imagesdata != 0)
	{
		grale::ImagesDataExtended *tmp = imagesdata;
		imagesdata = 0;
		if (!FindImagesData(tmp))
			delete tmp;
	}
	imagesdata = imgsdata;
}

bool GraleShellInstance::FindSource(grale::SourceImage *s)
{
	if (s == 0)
		return false;

	std::map<uint16_t, grale::SourceImage *>::const_iterator it;

	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		if ((*it).second == s)
			return true;
	}
	
	if (imgplane)
	{
		grale::SourcePlane *sp = imgplane->getSourcePlane();
		
		std::map<uint16_t, grale::SourceImage *> newsources = sp->getSources();
		std::map<uint16_t, grale::SourceImage *>::const_iterator it;

		for (it = newsources.begin() ; it != newsources.end() ; it++)
		{
			if ((*it).second == s)
				return true;
		}
	}
	return false;
}

bool GraleShellInstance::FindImagesData(grale::ImagesDataExtended *i)
{
	if (i == 0)
		return false;

	if (i == imagesdata)
		return true;
	
	std::list<grale::ImagesDataExtended *>::iterator it;

	for (it = imagesdatalist.begin() ; it != imagesdatalist.end() ; it++)
	{
		if ((*it) == i)
			return true;
	}
	return false;
}

void GraleShellInstance::SetGrid(grale::Grid *g)
{
	if (grid)
		delete grid;
	grid = g;
}

void GraleShellInstance::SetGAParams(GraleShellGAParams *p)
{
	if (gaparams)
		delete gaparams;
	gaparams = p;
}

void GraleShellInstance::SetCosmology(grale::Cosmology *pCosmology)
{
	if (m_pCosmology)
		delete m_pCosmology;
	m_pCosmology = pCosmology;
}

void GraleShellInstance::SetOpenCLKernel(grale::OpenCLKernel *pCLKernel)
{
#ifdef GRALECONFIG_SUPPORT_OPENCL
	if (m_pOpenCLKernel)
		delete m_pOpenCLKernel;

	m_pOpenCLKernel = pCLKernel;
#endif // GRALECONFIG_SUPPORT_OPENCL
}

